#include <zephyr/device.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>

#include <string.h>

LOG_MODULE_REGISTER(app, LOG_LEVEL_DBG);

#define UART_DEVICE_NODE DT_CHOSEN(uart_echo)
#define UART_BAUD_RATE   DT_PROP(UART_DEVICE_NODE, current_speed)

static const struct device *const uart_dev = DEVICE_DT_GET(UART_DEVICE_NODE);

#define RX_TIMEOUT_US 5000
#define RX_TIMEOUT_BITS RX_TIMEOUT_US * UART_BAUD_RATE / 1000000
_Static_assert(RX_TIMEOUT_BITS < (1 << 10), "rx timeout too big for baud");

#define DMA_BUF_SIZE 64
static uint8_t dma_buffer_0[DMA_BUF_SIZE];
static uint8_t dma_buffer_1[DMA_BUF_SIZE];
static uint8_t *dma_current_ptr = dma_buffer_0;
static uint8_t *dma_reserve_ptr = dma_buffer_1;

#define TEST_BUFFER_SIZE 32
uint8_t test_buffer[TEST_BUFFER_SIZE];
static volatile int test_rx_cnt;

void on_rx_data_ready(struct uart_event_rx const *evt)
{
	LOG_HEXDUMP_DBG(evt->buf + evt->offset, evt->len, "RX data");

	// Copy as much data as will fit into the test buffer.
	int buffer_free_space = MAX(TEST_BUFFER_SIZE - test_rx_cnt, 0);
	int copy_cnt = MIN(evt->len, buffer_free_space);
	memcpy(&test_buffer[test_rx_cnt], evt->buf + evt->offset, copy_cnt);
	test_rx_cnt += evt->len;
}

void on_dma_buffer_request()
{
	int err = uart_rx_buf_rsp(uart_dev, dma_reserve_ptr, DMA_BUF_SIZE);
	if (err != 0) {
		LOG_ERR("unable to provide reserve buffer (%d)", err);
	}
}

void on_dma_buffer_released(struct uart_event_rx_buf const *evt)
{
	if (evt->buf != dma_current_ptr) {
		LOG_ERR("unexpected released buffer");
		return;
	}

	// Swap the buffer pointers.
	dma_current_ptr = dma_reserve_ptr;
	dma_reserve_ptr = evt->buf;
}

void on_rx_stopped(struct uart_event_rx_stop const *evt)
{
	// Handle any outstanding received data.
	on_rx_data_ready(&evt->data);

	if (evt->reason == UART_ERROR_FRAMING) {
		LOG_ERR("UART framing error");
	} else {
		LOG_ERR("UART stopped with reason %d", evt->reason);
	}
}

void on_uart_event(struct device const *, struct uart_event *evt, void *)
{
	switch (evt->type) {
	case UART_TX_DONE:
		break;

	case UART_TX_ABORTED:
		break;

	case UART_RX_RDY:
		on_rx_data_ready(&evt->data.rx);
		break;

	case UART_RX_BUF_REQUEST:
		on_dma_buffer_request();
		break;

	case UART_RX_BUF_RELEASED:
		on_dma_buffer_released(&evt->data.rx_buf);
		break;

	case UART_RX_DISABLED:
		break;

	case UART_RX_STOPPED:
		on_rx_stopped(&evt->data.rx_stop);
		break;
	}
}

void send_uart_blocking(const char *buf, int msg_len)
{
	int err = uart_tx(uart_dev, buf, msg_len, SYS_FOREVER_MS);
	if (err != 0) {
		LOG_ERR("error sending message (%d)", err);
	}

	// Block for the transmit duration.
	int tx_time_us = msg_len * 10 * 1000000 / UART_BAUD_RATE;
	k_busy_wait(tx_time_us);
}

bool test_uart(int delay_us)
{
	test_rx_cnt = 0;
	memset(test_buffer, 0, sizeof(test_buffer));

	// Send the first half of the message.
	const char msg[16] = "01234567ABCDEFGH";
	send_uart_blocking(&msg[0], 8);

	// Wait for the specified amount of time.
	k_busy_wait(delay_us);

	// Send the seconds half of the message.
	send_uart_blocking(&msg[8], 8);

	// Wait long enough so that everything should have come in.
	k_usleep(RX_TIMEOUT_US * 2);

	// Validate that the data is correct.
	bool pass = true;
	if (test_rx_cnt != 16) {
		LOG_ERR("unexpected number of bytes received: %d", test_rx_cnt);
		pass = false;
	} else if (memcmp(test_buffer, msg, 16) != 0) {
		LOG_ERR("got corrupted data");
		pass = false;
	}

	if (!pass) {
		LOG_HEXDUMP_ERR(test_buffer, test_rx_cnt, "test_buffer");
	}

	// Give some time between tests to allow logs to flush.
	k_msleep(pass ? 50 : 100);
	return pass;
}

int main(void)
{
	LOG_INF("starting async_uart_echo app with baud %d", UART_BAUD_RATE);

	if (!device_is_ready(uart_dev)) {
		LOG_ERR("UART device not found!");
		return 0;
	}

	int ret = uart_callback_set(uart_dev, &on_uart_event, NULL);

	if (ret < 0) {
		if (ret == -ENOTSUP) {
			LOG_ERR("Async UART API support not enabled\n");
		} else if (ret == -ENOSYS) {
			LOG_ERR("UART device does not support async API\n");
		} else {
			LOG_ERR("Error setting UART callback: %d\n", ret);
		}
		return 0;
	}
	ret = uart_rx_enable(uart_dev, dma_current_ptr, DMA_BUF_SIZE, RX_TIMEOUT_US);

	int pass_cnt = 0;
	int total_cnt = 0;
	LOG_INF("sweeping UART delays");
	for (int delay_us = RX_TIMEOUT_US - 1000; delay_us < RX_TIMEOUT_US + 1000; delay_us += 10) {
		LOG_DBG("testing delay of %d us", delay_us);
		if (test_uart(delay_us)) {
			pass_cnt++;
		}

		total_cnt++;
	}

	LOG_INF("sweeping complete, %d/%d PASS", pass_cnt, total_cnt);

	while (true) {
		k_msleep(1000);
	}

	return 0;
}
